@@ -29,38 +29,20 @@
  * continuation).
  */
 
-static Parse_set * dummy_set(void)
-{
-	static Parse_set ds;
-	ds.first = ds.current = NULL;
-	ds.count = 1;
-	return &ds;
-}
-
-/** Returns an empty set of parses */
-static inline Parse_set * empty_set(void)
-{
-	Parse_set *s;
-	s = (Parse_set *) xalloc(sizeof(Parse_set));
-	s->first = s->current = NULL;
-	s->count = 0;
-	return s;
-}
-
 static void free_set(Parse_set *s)
 {
 	Parse_choice *p, *xp;
 	if (s == NULL) return;
-	for (p=s->first; p != NULL; p = xp) {
+	for (p=s->first; p != NULL; p = xp)
+	{
 		xp = p->next;
 		xfree((void *)p, sizeof(*p));
 	}
-	xfree((void *)s, sizeof(*s));
 }
 
 static Parse_choice *
-make_choice(Parse_set *lset, int llw, int lrw, Connector * llc, Connector * lrc,
-            Parse_set *rset, int rlw, int rrw, Connector * rlc, Connector * rrc,
+make_choice(Parse_set *lset, Connector * llc, Connector * lrc,
+            Parse_set *rset, Connector * rlc, Connector * rrc,
             Disjunct *ld, Disjunct *md, Disjunct *rd)
 {
 	Parse_choice *pc;
@@ -68,14 +50,14 @@ make_choice(Parse_set *lset, int llw, int lrw, Connector * llc, Connector * lrc,
 	pc->next = NULL;
 	pc->set[0] = lset;
 	pc->link[0].link_name = NULL;
-	pc->link[0].lw = llw;
-	pc->link[0].rw = lrw;
+	pc->link[0].lw = lset->lw;
+	pc->link[0].rw = lset->rw;
 	pc->link[0].lc = llc;
 	pc->link[0].rc = lrc;
 	pc->set[1] = rset;
 	pc->link[1].link_name = NULL;
-	pc->link[1].lw = rlw;
-	pc->link[1].rw = rrw;
+	pc->link[1].lw = rset->lw;
+	pc->link[1].rw = rset->rw;
 	pc->link[1].lc = rlc;
 	pc->link[1].rc = rrc;
 	pc->ld = ld;
@@ -85,7 +67,7 @@ make_choice(Parse_set *lset, int llw, int lrw, Connector * llc, Connector * lrc,
 }
 
 /**
- * Put this parse_choice into a given set.  The current pointer is always
+ * Put this parse_choice into a given set.  The tail pointer is always
  * left pointing to the end of the list.
  */
 static void put_choice_in_set(Parse_set *s, Parse_choice *pc)
@@ -96,12 +78,22 @@ static void put_choice_in_set(Parse_set *s, Parse_choice *pc)
 	}
 	else
 	{
-		s->current->next = pc;
+		s->tail->next = pc;
 	}
-	s->current = pc;
+	s->tail = pc;
 	pc->next = NULL;
 }
 
+static void record_choice(
+    Parse_set *lset, Connector * llc, Connector * lrc,
+    Parse_set *rset, Connector * rlc, Connector * rrc,
+    Disjunct *ld, Disjunct *md, Disjunct *rd, Parse_set *s)
+{
+	put_choice_in_set(s, make_choice(lset, llc, lrc,
+	                                 rset, rlc, rrc,
+	                                 ld, md, rd));
+}
+
 /**
  * Allocate the parse info struct
  *
@@ -118,8 +110,6 @@ Parse_info parse_info_new(int nwords)
 	memset(pi, 0, sizeof(struct Parse_info_struct));
 	pi->N_words = nwords;
 
-	pi->parse_set = NULL;
-
 	/* Alloc the x_table */
 	if (nwords >= 10) {
 		log2_table_size = 14;
@@ -151,10 +141,10 @@ void free_parse_info(Parse_info pi)
 
 	for (i=0; i<pi->x_table_size; i++)
 	{
-		for(t = pi->x_table[i]; t!= NULL; t=x)
+		for (t = pi->x_table[i]; t!= NULL; t=x)
 		{
 			x = t->next;
-			free_set(t->set);
+			free_set(&t->set);
 			xfree((void *) t, sizeof(X_table_connector));
 		}
 	}
@@ -176,11 +166,11 @@ static X_table_connector * x_table_pointer(int lw, int rw,
                               unsigned int null_count, Parse_info pi)
 {
 	X_table_connector *t;
-	t = pi->x_table[pair_hash(pi->log2_x_table_size, lw, rw, le, re, null_count)];
+	t = pi->x_table[pair_hash(pi->x_table_size, lw, rw, le, re, null_count)];
 	for (; t != NULL; t = t->next) {
-		if ((t->lw == lw) && (t->rw == rw) &&
-		    (t->le == le) && (t->re == re) &&
-		    (t->null_count == null_count))  return t;
+		if ((t->set.lw == lw) && (t->set.rw == rw) &&
+		    (t->set.le == le) && (t->set.re == re) &&
+		    (t->set.null_count == null_count))  return t;
 	}
 	return NULL;
 }
@@ -196,27 +186,75 @@ static X_table_connector * x_table_store(int lw, int rw,
 	unsigned int h;
 
 	n = (X_table_connector *) xalloc(sizeof(X_table_connector));
-	n->set = empty_set();
-	n->lw = lw; n->rw = rw; n->le = le; n->re = re; n->null_count = null_count;
-	h = pair_hash(pi->log2_x_table_size, lw, rw, le, re, null_count);
+	n->set.lw = lw;
+	n->set.rw = rw;
+	n->set.null_count = null_count;
+	n->set.le = le;
+	n->set.re = re;
+	n->set.count = 0;
+	n->set.first = NULL;
+	n->set.tail = NULL;
+
+	h = pair_hash(pi->x_table_size, lw, rw, le, re, null_count);
 	t = pi->x_table[h];
 	n->next = t;
 	pi->x_table[h] = n;
 	return n;
 }
 
-#ifdef UNUSED_FUNCTION
-static void x_table_update(int lw, int rw, Connector *le, Connector *re,
-                unsigned int null_count, Parse_set * set, Parse_info pi)
+/** Create a bogus parse set that only holds lw, rw. */
+static Parse_set* dummy_set(int lw, int rw,
+                            unsigned int null_count, Parse_info pi)
 {
-	/* Stores the value in the x_table.  Unlike x_table_store, it assumes it's already there */
-	X_table_connector *t = x_table_pointer(lw, rw, le, re, null_count, pi);
+	X_table_connector *dummy;
+	dummy = x_table_pointer(lw, rw, NULL, NULL, null_count, pi);
+	if (dummy) return &dummy->set;
 
-	assert(t != NULL, "This entry is supposed to be in the x_table.");
-	t->set = set;
+	dummy = x_table_store(lw, rw, NULL, NULL, null_count, pi);
+	dummy->set.count = 1;
+	return &dummy->set;
 }
-#endif
 
+#ifdef FINISH_THIS_IDEA_MAYBE_LATER
+static int cost_compare(const void *a, const void *b)
+{
+	const Match_node* const * ma = a;
+	const Match_node* const * mb = b;
+	if ((*ma)->d->cost < (*mb)->d->cost) return -1;
+	if ((*ma)->d->cost > (*mb)->d->cost) return 1;
+	return 0;
+}
+
+/**
+ * Sort the matchlist into ascending disjunct cost. The goal here
+ * is to issue the lowest-cost disjuncts first, so that the parse
+ * set ends up quasi-sorted. This is not enough to get us a totally
+ * sorted parse set, but it does guarantee that at least the very
+ * first parse really will be the lowest cost.
+ */
+static Match_node* sort_matchlist(Match_node* mlist)
+{
+	Match_node* mx;
+	Match_node** marr;
+	size_t len = 1;
+	size_t i;
+
+	for (mx = mlist; mx->next != NULL; mx = mx->next) len++;
+	if (1 == len) return mlist;
+
+	/* Avoid blowing out the stack. Its hopless. */
+	if (100000 < len) return mlist;
+
+	marr = alloca(len * sizeof(Match_node*));
+	i = 0;
+	for (mx = mlist; mx != NULL; mx = mx->next) marr[i++] = mx;
+
+	qsort((void *) marr, len, sizeof(Match_node*), cost_compare);
+	for (i=0; i<len-1; i++) marr[i]->next = marr[i+1];
+	marr[len-1]->next = NULL;
+	return marr[0];
+}
+#endif /* FINISH_THIS_IDEA_MAYBE_LATER */
 
 /**
  * returns NULL if there are no ways to parse, or returns a pointer
@@ -235,74 +273,81 @@ Parse_set * mk_parse_set(Sentence sent, fast_matcher_t *mchxt,
                  Connector *le, Connector *re, unsigned int null_count,
                  bool islands_ok, Parse_info pi)
 {
-	Disjunct * d, * dis;
 	int start_word, end_word, w;
-	bool Lmatch, Rmatch;
-	unsigned int lnull_count, rnull_count;
-	int i, j;
-	Parse_set *ls[4], *rs[4], *lset, *rset;
-	Parse_choice * a_choice;
-
-	Match_node * m, *m1;
 	X_table_connector *xt;
-	s64 count;
+	Count_bin * count;
 
 	assert(null_count < 0x7fff, "mk_parse_set() called with null_count < 0.");
 
 	count = table_lookup(ctxt, lw, rw, le, re, null_count);
 
-	/*
-	  assert(count >= 0, "mk_parse_set() called on params that were not in the table.");
-	  Actually, we can't assert this, because of the pseudocount technique that's
-	  used in count().  It's not the case that every call to mk_parse_set() has already
-	  been put into the table.
-	 */
-
-	if ((count == 0) || (count == -1)) return NULL;
+	/* If there's no counter, then there's no way to parse. */
+	if (NULL == count) return NULL;
+	if (hist_total(count) == 0) return NULL;
 
 	xt = x_table_pointer(lw, rw, le, re, null_count, pi);
 
-	if (xt != NULL) return xt->set;  /* we've already computed it */
+	/* Perhaps we've already computed it; if so, return it. */
+	if (xt != NULL) return &xt->set;
 
-	/* Start it out with the empty set of options. */
+	/* Start it out with the empty set of parse chocies. */
 	/* This entry must be updated before we return. */
 	xt = x_table_store(lw, rw, le, re, null_count, pi);
 
-	xt->set->count = count;  /* the count we already computed */
-	/* this count is non-zero */
+	/* The count we previously computed; its non-zero. */
+	xt->set.count = hist_total(count);
 
-	if (rw == 1 + lw) return xt->set;
+#define NUM_PARSES 4
+	// xt->set.cost_cutoff = hist_cost_cutoff(count, NUM_PARSES);
+	// xt->set.cut_count = hist_cut_total(count, NUM_PARSES);
 
+#define RECOUNT(X)  /* Make it disappear... */
+	RECOUNT({xt->set.recount = 1;})
+
+	/* If the two words are next to each other, the count == 1 */
+	if (lw + 1 == rw) return &xt->set;
+
+	/* The left and right connectors are null, but the two words are
+	 * NOT next to each-other.  */
 	if ((le == NULL) && (re == NULL))
 	{
-		if (!islands_ok && (lw != -1)) return xt->set;
+		Parse_set* pset;
+		Parse_set* dummy;
+		Disjunct* dis;
 
-		if (null_count == 0) return xt->set;
+		if (!islands_ok && (lw != -1)) return &xt->set;
+		if (null_count == 0) return &xt->set;
+
+		RECOUNT({xt->set.recount = 0;})
 
 		w = lw + 1;
 		for (dis = sent->word[w].d; dis != NULL; dis = dis->next)
 		{
 			if (dis->left == NULL)
 			{
-				rs[0] = mk_parse_set(sent, mchxt, ctxt, dis, NULL, w, rw, dis->right,
-				                  NULL, null_count-1, islands_ok, pi);
-				if (rs[0] == NULL) continue;
-				a_choice = make_choice(dummy_set(), lw, w, NULL, NULL,
-				                       rs[0], w, rw, NULL, NULL,
-				                       NULL, NULL, NULL);
-				put_choice_in_set(xt->set, a_choice);
+				pset = mk_parse_set(sent, mchxt, ctxt,
+				                    dis, NULL, w, rw, dis->right, NULL,
+				                    null_count-1, islands_ok, pi);
+				if (pset == NULL) continue;
+				dummy = dummy_set(lw, w, null_count-1, pi);
+				record_choice(dummy, NULL, NULL,
+				              pset,  NULL, NULL,
+				              NULL, NULL, NULL, &xt->set);
+				RECOUNT({xt->set.recount += pset->recount;})
 			}
 		}
-		rs[0] = mk_parse_set(sent, mchxt, ctxt, NULL, NULL, w, rw, NULL, NULL,
-		                  null_count-1, islands_ok, pi);
-		if (rs[0] != NULL)
+		pset = mk_parse_set(sent, mchxt, ctxt,
+		                    NULL, NULL, w, rw, NULL, NULL,
+		                    null_count-1, islands_ok, pi);
+		if (pset != NULL)
 		{
-			a_choice = make_choice(dummy_set(), lw, w, NULL, NULL,
-			                       rs[0], w, rw, NULL, NULL,
-			                       NULL, NULL, NULL);
-			put_choice_in_set(xt->set, a_choice);
+			dummy = dummy_set(lw, w, null_count-1, pi);
+			record_choice(dummy, NULL, NULL,
+			              pset,  NULL, NULL,
+			              NULL, NULL, NULL, &xt->set);
+			RECOUNT({xt->set.recount += pset->recount;})
 		}
-		return xt->set;
+		return &xt->set;
 	}
 
 	if (le == NULL)
@@ -323,63 +368,106 @@ Parse_set * mk_parse_set(Sentence sent, fast_matcher_t *mchxt,
 		end_word = re->word + 1;
 	}
 
-	/* This condition can never be true here. It is included so GCC will be able
-	 * to optimize the loop over "null_count".  Without this check, GCC thinks this
-	 * loop may be an infinite loop and it may omit some optimizations. */
+	/* This condition can never be true here. It is included so GCC
+	 * will be able to optimize the loop over "null_count".  Without
+	 * this check, GCC thinks this loop may be an infinite loop and
+	 * it may omit some optimizations. */
 	if (UINT_MAX == null_count) return NULL;
 
+	RECOUNT({xt->set.recount = 0;})
 	for (w = start_word; w < end_word; w++)
 	{
-		m1 = m = form_match_list(mchxt, w, le, lw, re, rw);
-		for (; m!=NULL; m=m->next)
+		Match_node * m, *mlist;
+		mlist = form_match_list(mchxt, w, le, lw, re, rw);
+		// if (mlist) mlist = sort_matchlist(mlist);
+		for (m = mlist; m != NULL; m = m->next)
 		{
-			d = m->d;
+			unsigned int lnull_count, rnull_count;
+			Disjunct* d = m->d;
 			for (lnull_count = 0; lnull_count <= null_count; lnull_count++)
 			{
-				rnull_count = null_count-lnull_count;
-				/* now lnull_count and rnull_count are the null_counts we're assigning to
-				 * those parts respectively */
+				int i, j;
+				bool Lmatch, Rmatch;
+				Parse_set *ls[4], *rs[4];
+
+				/* Here, lnull_count and rnull_count are the null_counts
+				 * we're assigning to those parts respectively. */
+				rnull_count = null_count - lnull_count;
 
 				/* Now, we determine if (based on table only) we can see that
 				   the current range is not parsable. */
 
-				Lmatch = (le != NULL) && (d->left != NULL) && do_match(le, d->left, lw, w);
-				Rmatch = (d->right != NULL) && (re != NULL) && do_match(d->right, re, w, rw);
-				for (i=0; i<4; i++) {ls[i] = rs[i] = NULL;}
+				Lmatch = (le != NULL) && (d->left != NULL)
+				         && do_match(le, d->left, lw, w);
+				Rmatch = (d->right != NULL) && (re != NULL)
+				         && do_match(d->right, re, w, rw);
+
+				for (i=0; i<4; i++) { ls[i] = rs[i] = NULL; }
 				if (Lmatch)
 				{
-					ls[0] = mk_parse_set(sent, mchxt, ctxt, ld, d, lw, w, le->next, d->left->next, lnull_count, islands_ok, pi);
-					if (le->multi) ls[1] = mk_parse_set(sent, mchxt, ctxt, ld, d, lw, w, le, d->left->next, lnull_count, islands_ok, pi);
-					if (d->left->multi) ls[2] = mk_parse_set(sent, mchxt, ctxt, ld, d, lw, w, le->next, d->left, lnull_count, islands_ok, pi);
-					if (le->multi && d->left->multi) ls[3] = mk_parse_set(sent, mchxt, ctxt, ld, d, lw, w, le, d->left, lnull_count, islands_ok, pi);
+					ls[0] = mk_parse_set(sent, mchxt, ctxt,
+					             ld, d, lw, w, le->next, d->left->next,
+					             lnull_count, islands_ok, pi);
+
+					if (le->multi)
+						ls[1] = mk_parse_set(sent, mchxt, ctxt,
+						              ld, d, lw, w, le, d->left->next,
+						              lnull_count, islands_ok, pi);
+
+					if (d->left->multi)
+						ls[2] = mk_parse_set(sent, mchxt, ctxt,
+						              ld, d, lw, w, le->next, d->left,
+						              lnull_count, islands_ok, pi);
+
+					if (le->multi && d->left->multi)
+						ls[3] = mk_parse_set(sent, mchxt, ctxt,
+						              ld, d, lw, w, le, d->left,
+						              lnull_count, islands_ok, pi);
 				}
+
 				if (Rmatch)
 				{
-					rs[0] = mk_parse_set(sent, mchxt, ctxt, d, rd, w, rw, d->right->next, re->next, rnull_count, islands_ok, pi);
-					if (d->right->multi) rs[1] = mk_parse_set(sent, mchxt, ctxt, d, rd, w,rw,d->right,re->next, rnull_count, islands_ok, pi);
-					if (re->multi) rs[2] = mk_parse_set(sent, mchxt, ctxt, d, rd, w, rw, d->right->next, re, rnull_count, islands_ok, pi);
-					if (d->right->multi && re->multi) rs[3] = mk_parse_set(sent, mchxt, ctxt, d, rd, w, rw, d->right, re, rnull_count, islands_ok, pi);
+					rs[0] = mk_parse_set(sent, mchxt, ctxt,
+					                 d, rd, w, rw, d->right->next, re->next,
+					                 rnull_count, islands_ok, pi);
+
+					if (d->right->multi)
+						rs[1] = mk_parse_set(sent, mchxt, ctxt,
+					                 d, rd, w, rw, d->right, re->next,
+						              rnull_count, islands_ok, pi);
+
+					if (re->multi)
+						rs[2] = mk_parse_set(sent, mchxt, ctxt,
+						              d, rd, w, rw, d->right->next, re,
+						              rnull_count, islands_ok, pi);
+
+					if (d->right->multi && re->multi)
+						rs[3] = mk_parse_set(sent, mchxt, ctxt,
+						              d, rd, w, rw, d->right, re,
+						              rnull_count, islands_ok, pi);
 				}
 
 				for (i=0; i<4; i++)
 				{
-					/* this ordering is probably not consistent with that
-					 *  needed to use list_links */
+					/* This ordering is probably not consistent with that
+					 * needed to use list_links. (??) */
 					if (ls[i] == NULL) continue;
 					for (j=0; j<4; j++)
 					{
 						if (rs[j] == NULL) continue;
-						a_choice = make_choice(ls[i], lw, w, le, d->left,
-						                       rs[j], w, rw, d->right, re,
-						                       ld, d, rd);
-						put_choice_in_set(xt->set, a_choice);
+						record_choice(ls[i], le, d->left,
+						              rs[j], d->right, re,
+						              ld, d, rd, &xt->set);
+						RECOUNT({xt->set.recount += ls[i]->recount * rs[j]->recount;})
 					}
 				}
 
 				if (ls[0] != NULL || ls[1] != NULL || ls[2] != NULL || ls[3] != NULL)
 				{
-					/* evaluate using the left match, but not the right */
-					rset = mk_parse_set(sent, mchxt, ctxt, d, rd, w, rw, d->right, re, rnull_count, islands_ok, pi);
+					/* Evaluate using the left match, but not the right */
+					Parse_set* rset = mk_parse_set(sent, mchxt, ctxt,
+					                        d, rd, w, rw, d->right, re,
+					                        rnull_count, islands_ok, pi);
 					if (rset != NULL)
 					{
 						for (i=0; i<4; i++)
@@ -387,48 +475,49 @@ Parse_set * mk_parse_set(Sentence sent, fast_matcher_t *mchxt,
 							if (ls[i] == NULL) continue;
 							/* this ordering is probably not consistent with
 							 * that needed to use list_links */
-							a_choice = make_choice(ls[i], lw, w, le, d->left,
-							         rset, w, rw, NULL /* d->right */,
-							         re,  /* the NULL indicates no link*/
-							         ld, d, rd);
-							put_choice_in_set(xt->set, a_choice);
+							record_choice(ls[i], le, d->left,
+							              rset,  NULL /* d->right */,
+							              re,  /* the NULL indicates no link*/
+							              ld, d, rd, &xt->set);
+							RECOUNT({xt->set.recount += ls[i]->recount * rset->recount;})
 						}
 					}
 				}
 				if ((le == NULL) && (rs[0] != NULL ||
 				     rs[1] != NULL || rs[2] != NULL || rs[3] != NULL))
 				{
-					/* evaluate using the right match, but not the left */
-					lset = mk_parse_set(sent, mchxt, ctxt, ld, d, lw, w, le, d->left, lnull_count, islands_ok, pi);
+					/* Evaluate using the right match, but not the left */
+					Parse_set* lset = mk_parse_set(sent, mchxt, ctxt,
+					                        ld, d, lw, w, le, d->left,
+					                        lnull_count, islands_ok, pi);
 
 					if (lset != NULL)
 					{
-						for (i=0; i<4; i++)
+						for (j=0; j<4; j++)
 						{
-							if (rs[i] == NULL) continue;
+							if (rs[j] == NULL) continue;
 							/* this ordering is probably not consistent with
 							 * that needed to use list_links */
-							a_choice = make_choice(lset, lw, w, NULL /* le */,
-							          d->left,  /* NULL indicates no link */
-							          rs[i], w, rw, d->right, re,
-							          ld, d, rd);
-							put_choice_in_set(xt->set, a_choice);
+							record_choice(lset, NULL /* le */,
+							                    d->left,  /* NULL indicates no link */
+							              rs[j], d->right, re,
+							              ld, d, rd, &xt->set);
+							RECOUNT({xt->set.recount += lset->recount * rs[j]->recount;})
 						}
 					}
 				}
 			}
 		}
-		put_match_list(mchxt, m1);
+		put_match_list(mchxt, mlist);
 	}
-	xt->set->current = xt->set->first;
-	return xt->set;
+	return &xt->set;
 }
 
 /**
- * return TRUE if and only if overflow in the number of parses occurred.
- * Use a 64-bit int for counting.
+ * Return TRUE if and only if an overflow in the number of parses
+ * occurred. Use a 64-bit int for counting.
  */
-static bool verify_set_node(Parse_set *set)
+static bool set_node_overflowed(Parse_set *set)
 {
 	Parse_choice *pc;
 	s64 n = 0;
@@ -442,17 +531,17 @@ static bool verify_set_node(Parse_set *set)
 	return false;
 }
 
-static bool verify_set(Parse_info pi)
+static bool set_overflowed(Parse_info pi)
 {
 	unsigned int i;
 
-	assert(pi->x_table != NULL, "called verify_set when x_table==NULL");
+	assert(pi->x_table != NULL, "called set_verflowed with x_table==NULL");
 	for (i=0; i<pi->x_table_size; i++)
 	{
 		X_table_connector *t;
-		for(t = pi->x_table[i]; t != NULL; t = t->next)
+		for (t = pi->x_table[i]; t != NULL; t = t->next)
 		{
-			if (verify_set_node(t->set)) return true;
+			if (set_node_overflowed(&t->set)) return true;
 		}
 	}
 	return false;
@@ -477,20 +566,13 @@ bool build_parse_set(Sentence sent, fast_matcher_t *mchxt,
                     count_context_t *ctxt,
                     unsigned int null_count, Parse_Options opts)
 {
-	Parse_set * whole_set;
-
-	whole_set =
+	sent->parse_info->parse_set =
 		mk_parse_set(sent, mchxt, ctxt,
 		             NULL, NULL, -1, sent->length, NULL, NULL, null_count+1,
 		             opts->islands_ok, sent->parse_info);
 
-	if ((whole_set != NULL) && (whole_set->current != NULL)) {
-		whole_set->current = whole_set->first;
-	}
 
-	sent->parse_info->parse_set = whole_set;
-
-	return verify_set(sent->parse_info);
+	return set_overflowed(sent->parse_info);
 }
 
 static void issue_link(Linkage lkg, Disjunct * ld, Disjunct * rd, Link * link)
