@@ -20,12 +20,21 @@ struct PositionConnector
 {
   PositionConnector(Exp* e, Connector* c, char d, int w, int p, 
                     double cst, double pcst, bool lr, bool ll,
-                    const std::vector<int>& er, const std::vector<int>& el)
-    : exp(e), connector(c), dir(d), word(w), position(p),
+                    const std::vector<int>& er, const std::vector<int>& el, const X_node *w_xnode)
+    : exp(e), dir(d), word(w), position(p),
       cost(cst), parent_cost(pcst),
       leading_right(lr), leading_left(ll),
-      eps_right(er), eps_left(el)
+      eps_right(er), eps_left(el), word_xnode(w_xnode)
   {
+    // Initialize some fields in the connector struct.
+    connector.string = c->string;
+    connector.multi = c->multi;
+    connector.length_limit = c->length_limit;
+
+    if (word_xnode == NULL) {
+       cerr << "Internal error: Word" << w << ": " << "; connector: '" << c->string << "'; X_node: " << (word_xnode?word_xnode->string: "(null)") << endl;
+    }
+
     /*
     cout << c->string << " : ." << w << ". : ." << p << ". ";
     if (leading_right) {
@@ -44,7 +53,7 @@ struct PositionConnector
   Exp* exp;
 
   // Connector itself
-  Connector* connector;
+  Connector connector;
   // Direction
   char dir;
   // word in a sentence that this connector belongs to
@@ -61,6 +70,10 @@ struct PositionConnector
   std::vector<int> eps_right;
   std::vector<int> eps_left;
 
+
+  // The corresponding X_node - chosen-disjuncts[]
+  const X_node *word_xnode;
+
   // Matches with other words
   std::vector<PositionConnector*> matches;
 
@@ -95,6 +108,9 @@ public:
   WordTag(int word, Variables* variables, Sentence sent, Parse_Options opts)
     : _word(word), _variables(variables), _sent(sent), _opts(opts) {
     _match_possible.resize(_sent->length);
+    verbosity = opts->verbosity;
+    debug = opts->debug;
+    test = opts->test;
   }
 
   const std::vector<PositionConnector>& get_left_connectors() const {
@@ -118,18 +134,16 @@ public:
 
   void set_connector_length_limit(Connector* c)
   {
-    int short_len = _opts->short_length;
-    if (short_len > UNLIMITED_LEN)
-      short_len = UNLIMITED_LEN;
+    unsigned int len = _opts->short_length;
+    Connector_set * conset = _sent->dict->unlimited_connector_set;
 
-    Connector_set *conset = _sent->dict->unlimited_connector_set;
-    if (parse_options_get_all_short_connectors(_opts)) {
-      c->length_limit = short_len;
-    }
-    else if (conset == NULL || match_in_connector_set(conset, c, '+')) {
+    if (len > UNLIMITED_LEN) len = UNLIMITED_LEN;
+
+    if (!_opts->all_short &&
+        (conset == NULL || match_in_connector_set(conset, c))) {
       c->length_limit = UNLIMITED_LEN;
     } else {
-      c->length_limit = short_len;
+      c->length_limit = len;
     }
   }
 
@@ -146,7 +160,7 @@ public:
                          std::vector<int>& eps_right,
                          std::vector<int>& eps_left,
                          char* var, bool root, double parent_cost,
-                         Exp* parent);
+                         Exp* parent, const X_node *word_xnode);
 
   // Caches information about the found matches to the _matches vector, and also
   // updates the _matches vector of all connectors in the given tag.
@@ -165,6 +179,10 @@ public:
     return _match_possible[wi].find(pi) != _match_possible[wi].end();
   }
 
+private:
+  int verbosity;
+  const char *debug;
+  const char *test;
 };
 
 #endif
